/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */
package isql;

import isql.array.Array;
import isql.builder.ExpressionBuilder;
import isql.builder.SelectExpressionBuilder;
import isql.builder.WhereBuilder;
import isql.groupby.Groups;
import isql.predicate.WhereView;
import isql.util.ArrayBuilder;
import isql.util.ILogger;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import isql.ColumnType;

/**
 * Builds the structures to execute the SQL query for all the SQL query parts 
 * (SELECT, WHERE, GROUP BY...) 
 * @author SHZ Dec 6, 2007
 */
public class ClausesBuilder
{
	/**
	 * To log what is going on when building the structures
	 */
	private ILogger log;
	/**
	 * Constructor
	 * @param log the logger
	 */
	public ClausesBuilder(ILogger log)
	{
		this.log = log;
	}
	
	/**
	 * Builds a Groups structure (containing the group by relationships) from the grammars
	 * output
	 * @param lstGroupByTableColumn the columns part of the GROUP BY clause (from ANTLR)
	 * @param values the result set at this point in the query evaluation 
	 * @return the Groups structure 
	 * @throws ISQLException
	 */
	public Groups buildGroupBy(List<ISQLSelectParser.TableColumn> lstGroupByTableColumn, ColumnValuesWithRowsNumber values) throws ISQLException
	{
		if(lstGroupByTableColumn.size() > 0)
		{
			List<ColumnKey> lstGroupByKey = new ArrayList<ColumnKey>();
			for(ISQLSelectParser.TableColumn tableColumn : lstGroupByTableColumn)
				lstGroupByKey.add(ExpressionBuilder.buildTableColumn(tableColumn));
			GroupBy groupBy = new GroupBy(lstGroupByKey);
			return groupBy.apply(values);
		}
		else
			return null;
	}



	/**
	 * Builds a map alias -> result expression from the select clause.
	 * To do that it applies the GROUP BY clause to the current result set
	 * @param lstSelect the list of select items
	 * @param values the result set at this point of the query evaluation
	 * @param groups the structure defining the groups of the GROUP BY clause 
	 * @return the map alias -> result expression
	 * @throws ISQLException
	 */
	public Map<String, IExpression> buildSelect(List<ISQLSelectParser.SelectItem> lstSelect, ColumnValues values, Groups groups) throws ISQLException
	{
		ExpressionBuilder builder = new SelectExpressionBuilder(values, groups);
		Map<String, IExpression> mapSelect = new TreeMap<String, IExpression>();
		try
		{
			for(ISQLSelectParser.SelectItem item : lstSelect)
			{
				IExpression expression = builder.buildExpression(item.getWhat());
				mapSelect.put(item.getAlias(), expression);
			}
			return mapSelect;
		}
		catch (ISQLException e)
		{
			throw new ISQLException("Building the select items got ", e);
		}		
	}

	/**
	 * from the result of the where clause (where view) and the original column values
	 * builds a new set of column values
	 * @param values the column in the tables with their values. Since they are
	 * 	of different tables, they can still have different size
	 * @param whereView the view resulting from the application of the where clause
	 * to the column values
	 * @return the result set after the where
	 * @throws ISQLException 
	 */
	public ColumnValuesWithRowsNumber buildViewAfterWhere(ColumnValues values, WhereView whereView) throws ISQLException
	{
		Map<ColumnKey, Array> mapValuesByColumnAfterWhere = new TreeMap<ColumnKey, Array>();
		for(Map.Entry<ColumnKey, Array> entry : values.entrySet())
		{
			ColumnKey key = entry.getKey();
			Array initial = entry.getValue();
			ColumnType type = initial.getType();
			Object[] arrInitial = initial.getArray();
			List<Integer> lstIndex = whereView.getIndicesInTable(key.getTableName());
			int size = lstIndex.size();
			Object[] arrAfter = ArrayBuilder.buildArray(type, size);
			for(int i = 0; i < size; i++)
			{
				Integer index = lstIndex.get(i);
				arrAfter[i] = arrInitial[index.intValue()];
			}
			Array after = new Array(arrAfter, type);
			mapValuesByColumnAfterWhere.put(key, after);
		}
		return new ColumnValuesWithRowsNumber(mapValuesByColumnAfterWhere);
	}

	/**
	 * Returns the where clause in form of a predicate tree
	 * @param valuesOriginal the column in the tables with their values. Since they are
	 * 	of different tables, they can still have different size
	 * @param where the where condition extracted from ANTLR
	 * @return the resulting predicate tree 
	 * @throws ISQLException
	 */
	public IPredicate buildWhere(ColumnValues valuesOriginal, ISQLSelectParser.ISearchCondition where)
			throws ISQLException
	{
		WhereBuilder builder = new WhereBuilder(valuesOriginal);
	    try
		{
			if(where != null)				
				return builder.build(where);
			else
			{
				log.debug("No where clause");
				return null;
			}
		}
		catch (ISQLException e)
		{
			throw new ISQLException("Building the where predicate got ", e);
		}
	}

}
