/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */

package isql;

import isql.array.Array;
import isql.constant.BooleanConstant;
import isql.constant.DateConstant;
import isql.constant.NumericConstant;
import isql.constant.TextConstant;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Arrays;

import isql.ColumnType;

/**
 * The parts of an expression or the expression
 * itself can be either arrays or constants,
 * This is a problem, because to calculate the expression we
 * need a set of array of the same size.
 * To do that we need to know the size of all the arrays involved
 * (which must be the same for all) and build arrays from the constants
 * (with all the same values) with the same size of the other arrays.
 * The idea here is to use futures that get values when the equalize
 * method is called
 * 
 * @author SHZ Sep 28, 2007
 */
public class Equalizer
{
	/**
	 * Get the array, when it is ready (after
	 * the method equalize has been called)
	 * @author SHZ Sep 28, 2007
	 */
	public interface IArrayFuture
	{
		Array getArray() throws ISQLException;
	}
	
	
	/**
	 * The array can be created only when the size
	 * has been set
	 * @author SHZ Sep 28, 2007
	 */
	private static class ConstantFuture implements IArrayFuture
	{
		private IConstant value;
		private int size = 0;
		
		public ConstantFuture(IConstant value)
		{
			this.value = value;
		}
		
		public void setSize(int size)
		{
			this.size = size;
		}

		/* (non-Javadoc)
		 * @see isql.Equalizer.IArrayFuture#getArray()
		 */
		public Array getArray() throws ISQLException
		{
			if(size == 0)
				throw new ISQLException("Please call the equalized method");
			ColumnType type = value.getType();
			switch(type)
			{
			case NUMERIC:
			{
				Double[] ret = new Double[size];
				Arrays.fill(ret, ((NumericConstant)value).getValue());
				return new Array(ret, ColumnType.NUMERIC);
			}
			case TEXT:
			{
				String[] ret = new String[size];
				Arrays.fill(ret, ((TextConstant)value).getValue());
				return new Array(ret, ColumnType.TEXT);
			}
			case BOOLEAN:
			{
				Boolean[] ret = new Boolean[size];
				Arrays.fill(ret, ((BooleanConstant)value).getValue());
				return new Array(ret, ColumnType.BOOLEAN);
			}
			case DATE:
			{
				Timestamp[] ret = new Timestamp[size];
				Arrays.fill(ret, ((DateConstant)value).getValue());
				return new Array(ret, ColumnType.DATE);
			}
			default:
				throw new AssertionError("The type "+type+" has not been recogized");
			}
		}
	}
		
	/**
	 * The array here is ready from the start
	 * @author SHZ Sep 28, 2007
	 */
	private static class ArrayFuture implements IArrayFuture
	{
		private Array value;

		public ArrayFuture(Array value)
		{
			this.value = value;
		}

		/* (non-Javadoc)
		 * @see isql.Equalizer.IArrayFuture#getArray()
		 */
		public Array getArray()
		{
			return value;
		}
		
		public int getSize()
		{
			return value.getSize();
		}
		
	}
	
	/**
	 * The results of the expressions that contain an array
	 */
	private ArrayList<ArrayFuture> lstArray = new ArrayList<ArrayFuture>();
	/**
	 * The results of the expressions that contain a constant
	 */
	private ArrayList<ConstantFuture> lstConstant = new ArrayList<ConstantFuture>();
	/**
	 * Add a new expression. 
	 * @param value
	 * @return the future value of the expression. The idea is that you cannot access the real value of the expression
	 * until the equalize method is called.
	 */
	public IArrayFuture add(ITypedValue value)
	{	
		if(value instanceof IConstant)
		{
			ConstantFuture future = new ConstantFuture((IConstant) value);
			lstConstant.add(future);
			return future;
		}
		else if(value instanceof Array)
		{
			ArrayFuture future = new ArrayFuture((Array) value);
			lstArray.add(future);
			return future;
		}
		else
			throw new AssertionError("ITypedValue object "+value+" of class "+value.getClass()+" not recognized");
	}	
	/**
	 * Calculates the size of the constant arrays based on the size
	 * of the other arrays
	 * @throws ISQLException
	 */
	public void equalize() throws ISQLException
	{		
		int size;
		if(lstArray.size() > 0)
			size = getArraysSize();
		else
			size = 1; // only constants are involved. size 1 is then okay
		for(ConstantFuture future : lstConstant)
			future.setSize(size);
			
	}

	/**
	 * Finds out the size of all the arrays
	 * involved in the calculation. If the
	 * sizes of the arrays are different, an exception
	 * is raised 
	 * @throws ISQLException 
	 */
	private int getArraysSize() throws ISQLException
	{
		ArrayFuture future = lstArray.get(0);
		int size = future.getSize();
		for(int i = 1; i < lstArray.size(); i++)
		{
			future = lstArray.get(i);
			if(future.getSize() != size)
				throw new ISQLException("The arrays size is different"); 
		}
		return size;
	}
}
