/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */

package isql;

import isql.array.Array;
import isql.groupby.ColumnKeyAndType;
import isql.groupby.GroupKey;
import isql.groupby.Groups;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;


/**
 * Builds the group by structure.
 * The idea is that, after composed the so called "where view", which contains references to 
 * the rows of the tables are part of the query result, on that rows it
 * applies the group by clause, producing groups of rows. 
 * Each group is defined by the group by columns, for which the rows values are always the same.
 *   
 * @author SHZ Oct 2, 2007
 */
public class GroupBy
{
	/**
	 * The columns the compose the group by
	 */
	private Set<ColumnKey> setGroupByColumnKey;
	
	/**
	 * Constructor
	 * @param lstColumnKey the columns in the GROUP BY
	 */
	public GroupBy(List<ColumnKey> lstColumnKey)
	{
		this.setGroupByColumnKey = new TreeSet<ColumnKey>(lstColumnKey);
	}
	
	/**
	 * Divides the result set in sets of rows, one set of rows for each
	 * group. Each group is identified by the values of the GROUP BY columns, which are the same
	 * for all rows in the group.
	 * @param values the current result set
	 * @return the Groups structure, containing all the groups 
	 * @throws ISQLException
	 */
	public Groups apply(ColumnValuesWithRowsNumber values) throws ISQLException
	{
		ColumnKeyAndType[] arrGroupKeys = buildGroupKeysSameOrderAsColumns(values);
		return buildGroups(values, arrGroupKeys);		
	}

	/**
	 * returns an array of the column keys in the GROUP BY clause ordered
	 * in the way the same columns are ordered in the current result set 
	 * @param values the current result set
	 * @return the ordered array of group by columns
	 */
	private ColumnKeyAndType[] buildGroupKeysSameOrderAsColumns(
			ColumnValuesWithRowsNumber values)
	{
		// contains the columns that are part of the group by clause
		List<ColumnKeyAndType> lstGroupKeys = new ArrayList<ColumnKeyAndType>();
		for(Map.Entry<ColumnKey, Array> entry : values.entrySet())
		{
			ColumnKey key = entry.getKey();
			Array value = entry.getValue();
			// it is a group by column
			if(setGroupByColumnKey.contains(key))
				lstGroupKeys.add(new ColumnKeyAndType(key, value.getType()));
		}
		ColumnKeyAndType[] arrGroupKeys = lstGroupKeys.toArray(new ColumnKeyAndType[lstGroupKeys.size()]);
		return arrGroupKeys;
	}
	
	/**
	 * For each row of the result set extracts the GROUP BY columns values and
	 * with the forms a key, called group key, which stores in the resulting Groups structure
	 * together with the row index.
	 * @param values the current result set
	 * @param arrGroupKeys the columns in the GROUP BY clause, ordered in the same way
	 * as they are ordered in the result set
	 * @return the Group structure
	 */
	private Groups buildGroups(ColumnValuesWithRowsNumber values,
			ColumnKeyAndType[] arrGroupKeys)
	{
		Groups groups = new Groups(arrGroupKeys);
		int rows = values.getNumberOfRows();
		for(int i = 0; i < rows; i++)
		{
			Object[] arrValues = new Object[arrGroupKeys.length];
			for(int j = 0; j < arrGroupKeys.length; j++)
			{
				ColumnKeyAndType groupKey = arrGroupKeys[j];
				Array arrForGroupKey = values.get(groupKey.getKey());
				arrValues[j] = arrForGroupKey.getArray()[i];
			}
			groups.add(new GroupKey(arrValues), i);
		}
		return groups;
	}
	
}
