/*
 * Copyright (C) 2008 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LLGPL v3.0.
 */
package isql;

import isql.array.Array;
import isql.array.InputArray;
import isql.groupby.Groups;
import isql.predicate.WhereView;
import isql.util.CollectionsToString;
import isql.util.ILogger;

import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * The entry point to the ISQL library.
 * To make a query on arrays, create one instance of this class and call the query method 
 * @author SHZ Mar 25, 2008
 */
public class ISQLQuery
{
	/**
	 * To log the parsing and execution of the query
	 */
	private ILogger log;
	/**
	 * If this is active, a lot of debugging information is written in the log.
	 * Good for test purposes, but to remove in production.
	 */
	private boolean isDebug;
	/**
	 * Constructor
	 * @param isDebug if this is true, does some fine log about what is going on during a query
	 */
	public ISQLQuery(ILogger log, boolean isDebug)
	{
		this.log = log;
		this.isDebug = isDebug;
	}

	/**
	 * Builds the original map column key -> array from the input matrices
	 * @param mapArrayByName a map [column name -> array], with the column name identifying the array in the SQL query
	 * @param mapTableSource a map [table name -> table] with all the tables in the FROM clause of the SQL query (from ANTLR) 
	 * @return the arrays structured as table columns
	 * @throws ISQLException 
	 */
	private ColumnValues buildColumnValues(Map<String, InputArray> mapArrayByName, 
			Map<String, ISQLSelectParser.TableSource> mapTableSource)
			throws ISQLException
	{
		Map<ColumnKey, Array> mapValuesByColumnOriginal = new TreeMap<ColumnKey, Array>();
		for(ISQLSelectParser.TableSource tableSource : mapTableSource.values())
		{
			String tableName = tableSource.getTable();
			for(String columnName : (List<String>)tableSource.getColumns())
			{
				Array array = mapArrayByName.get(columnName);
				if(array == null)
					throw new ISQLException("Matrix with name "+columnName+" not found");
				mapValuesByColumnOriginal.put(new ColumnKey(tableName, columnName), array);
			}
		}
		return new ColumnValues(mapValuesByColumnOriginal);
	}



	/**
	 * Checks the initial size of the matrices/columns to avoid that two columns of different
	 * size end up in the same table
	 * @param values the columns with their array 
	 * @throws ISQLException 
	 */
	private void checkInitialSize(ColumnValues values) throws ISQLException
	{
		Map<String, Integer> mapSizeByTable = new TreeMap<String, Integer>();
		for(Map.Entry<ColumnKey, Array> entry : values.entrySet())
		{
			ColumnKey key = entry.getKey();
			String tableName = key.getTableName();
			Integer sizeForTable = mapSizeByTable.get(tableName);
			int sizeForColumn = entry.getValue().getSize(); 
			if(sizeForTable == null)
				mapSizeByTable.put(tableName, Integer.valueOf(sizeForColumn));
			else
			{
				if(sizeForTable.intValue() != sizeForColumn)
					throw new ISQLException("The columns of the table "+tableName+" do not have all the same size");
			}				
		}
	}

	/**
	 * Queries the arrays using the SQL query 
	 * @param sql the sql query, written in the ISQL sql dialect
	 * @param mapArrayByName a map [column name -> array], with the column name identifying the array in the SQL query
	 * @return a map [result name -> value] where the result name is the one that identifies each column in the result set of the SQL query (AS ...)
	 * 		and value can be either an array or a single value  
	 * @throws ISQLException errors during the query
	 */
	public Map<String, ITypedValue> query(String sql,
			Map<String, InputArray> mapArrayByName) throws ISQLException
	{
		SQLParser parser = new SQLParser(log);
		ISQLSelectParser g = parser.buildParser(sql);
		Map<String, ISQLSelectParser.TableSource> mapTableSource =  g.getTableSources();
		Map<String, IExpression> mapSelect = null;
		if(isDebug) log.debug("Build column values");
		ColumnValues valuesOriginal = buildColumnValues(mapArrayByName, mapTableSource);
		if(isDebug) log.debug("Check initial size");
		checkInitialSize(valuesOriginal);
		if(isDebug) log.debug("Build where");
		ClausesBuilder clausesBuilder = new ClausesBuilder(log);
		IPredicate wherePredicate = clausesBuilder.buildWhere(valuesOriginal, g.getWhere());        	
		WhereView whereView;
		if(wherePredicate != null)
			whereView = wherePredicate.apply();
		else
			whereView = WhereView.fromColumns(valuesOriginal);
		if(isDebug) log.debug("Build values after where");
		ColumnValuesWithRowsNumber valuesAfterWhere = clausesBuilder.buildViewAfterWhere(valuesOriginal, whereView);
		if(isDebug) log.debug("Values after where: "+valuesAfterWhere);			
		if(isDebug) log.debug("Build values after order by");
		ColumnValuesWithRowsNumber valuesAfterOrderBy = new OrderBy(log).applyOrderBy(g.getOrderBys(), valuesAfterWhere);
		if(isDebug) log.debug("Values after order by: "+valuesAfterOrderBy);
		if(isDebug) log.debug("Build group by");
		Groups groups = clausesBuilder.buildGroupBy(g.getGroupBys(), valuesAfterOrderBy);
		if(isDebug) log.debug("Group by: "+groups);
		if(isDebug) log.debug("Build select");
		mapSelect = clausesBuilder.buildSelect(g.getSelectItems(), valuesAfterOrderBy, groups);
		if(isDebug) log.debug("Select: "+CollectionsToString.mapToString(mapSelect));
		Map<String, ITypedValue> mapResult = convertMapExpressionToMapResult(mapSelect);				
		return mapResult;
	}

	/**
	 * The query returns a map name -> expression. This converts it
	 * to a map name -> result (array or single value), which is better
	 * as return value for the query 
	 * @param mapSelect the expressions in the select clause of the SQL query
	 * @return the result set values
	 * @throws ISQLException
	 */
	private Map<String, ITypedValue> convertMapExpressionToMapResult(
			Map<String, IExpression> mapSelect) throws ISQLException
	{
		Map<String, ITypedValue> mapResult = new TreeMap<String, ITypedValue>();
		for(Map.Entry<String, IExpression> entry : mapSelect.entrySet())
		{
			String name = entry.getKey();
			ITypedValue result = entry.getValue().getResult();
			mapResult.put(name, result);
		}
		return mapResult;
	}

}
