/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */
package isql;

import isql.array.Array;
import isql.builder.ExpressionBuilder;
import isql.orderby.OrderByKey;
import isql.orderby.OrderByKeyBehaviour;
import isql.util.ArrayBuilder;
import isql.util.ILogger;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import isql.ColumnType;
import isql.util.MapOfListHelp;

/**
 * Applies the order by clause to a set of columns/values
 * @author SHZ Nov 14, 2007
 */
public class OrderBy
{
	/**
	 * To log what happens applying the order by
	 */
	private ILogger log;
	/**
	 * Constructor 
	 * @param log the logger
	 */
	public OrderBy(ILogger log)
	{
		this.log = log;
	}
	/**
	 * Applies the ORDER BY clause
	 * @param lstOrderByColumn the columns in the ORDER BY condition
	 * @param valuesBefore the result set before the order by is applied 
	 * @return the result set after the order by
	 * @throws ISQLException
	 */
	public ColumnValuesWithRowsNumber applyOrderBy(List<ISQLSelectParser.OrderBy> lstOrderByColumn, ColumnValuesWithRowsNumber valuesBefore) throws ISQLException
	{
		if(lstOrderByColumn.size() > 0)
		{
			// builds a map column key -> array for all the columns that are included in the order by clause
			Map<ColumnKey, Array> mapOrderByKey = new TreeMap<ColumnKey, Array>();
			Map<ColumnKey, Boolean> mapOrderByAscendent = new TreeMap<ColumnKey, Boolean>();
			for(ISQLSelectParser.OrderBy orderBy : lstOrderByColumn)
			{
				ISQLSelectParser.TableColumn column = orderBy.getOrderBy();
				ColumnKey key = ExpressionBuilder.buildTableColumn(column);
				Array array = valuesBefore.get(key);				
				mapOrderByKey.put(key, array);
				mapOrderByAscendent.put(key, orderBy.isAscendent());
			}	
			OrderByKeyBehaviour behaviour = buildBehaviour(mapOrderByAscendent);
			Map<OrderByKey, List<Integer>> mapIndexBy = buildMapIndexByOrderKey(valuesBefore, behaviour, mapOrderByKey);
			// builds a new map with the same columns (in the same order) and the same number of rows of the original
			Map<ColumnKey, Array> mapValuesByColumnAfterOrderBy = new TreeMap<ColumnKey, Array>();
			int rows = valuesBefore.getNumberOfRows(); 
			for(Map.Entry<ColumnKey, Array> entry : valuesBefore.entrySet())
			{
				ColumnType type = entry.getValue().getType();
				mapValuesByColumnAfterOrderBy.put(entry.getKey(), new Array(ArrayBuilder.buildArray(type, rows), type));
			}
			// copies the rows of the original map to the final in the new order
			// Since the maps have the same column keys, they have the same number of elements and the same order-
			// Therefore we can use two iterators to assign one to the other
			int newRowIndex = 0;
			for(List<Integer> lstForOrderKey : mapIndexBy.values())
			{
				for(Integer originalRowIndex : lstForOrderKey)
				{
					Iterator<Array> iWhere = valuesBefore.values().iterator();
					Iterator<Array> iOrderBy = mapValuesByColumnAfterOrderBy.values().iterator();
					while(iWhere.hasNext())
					{
						Array arrWhere = iWhere.next();
						Array arrOrderBy = iOrderBy.next();
						arrOrderBy.getArray()[newRowIndex] = arrWhere.getArray()[originalRowIndex];
					}
					newRowIndex++;					
				}
			}
			return new ColumnValuesWithRowsNumber(mapValuesByColumnAfterOrderBy);
		}
		else
		{
			log.debug("No order by");			
			return valuesBefore;
		}
	}

	/**
	 * From a map [column key] -> [ascendent/descendent] builds the OrderBehaviour object
	 * that defines how the result set rows have to be sorted
	 * @param mapOrderByAscendent
	 * @return
	 */
	private OrderByKeyBehaviour buildBehaviour(Map<ColumnKey, Boolean> mapOrderByAscendent)
	{
		boolean[] arrAscendent = new boolean[mapOrderByAscendent.size()];
		int columnIndex = 0;
		for(Boolean ascendent : mapOrderByAscendent.values())
		{
			arrAscendent[columnIndex] = ascendent.booleanValue();
			columnIndex++;
		}
		OrderByKeyBehaviour behaviour = new OrderByKeyBehaviour(arrAscendent);
		return behaviour;
	}
	
	/**
	 * Builds a map [values of the order by columns in a row] -> [indexes of all the rows having these values
	 *  in the result set].
	 * The point here is that the same order by values can occur more than once in the result set
	 * so there can be more than one index for each row content.
	 * @param valuesBefore the result set at this point of the query evaluation
	 * @param behaviour how the ORDER BY orders the rows (ascendent...) 
	 * @param mapOrderByKey the ORDER BY clause organized as a map [column key] -> [column values]  
	 * @return the map [values of the order by columns in a row] -> [indexes of all the rows having these values
	 *  in the result set].
	 * @throws ISQLException
	 */
	private Map<OrderByKey, List<Integer>> buildMapIndexByOrderKey(ColumnValuesWithRowsNumber valuesBefore, OrderByKeyBehaviour behaviour, Map<ColumnKey, Array> mapOrderByKey) throws ISQLException
	{
		// builds a map order by key -> row index for each row of the select result 
		Map<OrderByKey, List<Integer>> mapIndexBy = new TreeMap<OrderByKey, List<Integer>>();
		int rows = valuesBefore.getNumberOfRows(); 
		int columns = mapOrderByKey.size();
		for(int i = 0; i < rows; i++)
		{
			// the values for a certain result set row
			Object[] arrKey = new Object[columns];
			int columnIndex = 0;
			for(Array array : mapOrderByKey.values())
			{
				arrKey[columnIndex] = array.getArray()[i];
				columnIndex++;
			}
			OrderByKey key = new OrderByKey(arrKey, behaviour);
			MapOfListHelp.put(mapIndexBy, key, Integer.valueOf(i));
		}
		return mapIndexBy;
	}

}
