/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */
package isql;

import isql.util.ILogger;

import java.util.List;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.MismatchedTokenException;
import org.antlr.runtime.RecognitionException;

/**
 * Uses ANTLR to parse the SQL queries and builds the error text if the SQL
 * parsing failed
 * @author SHZ Dec 7, 2007
 */
public class SQLParser
{
	/**
	 * To log what's going on during the SQL query parsing
	 */
	private ILogger log;
	/**
	 * Constructor
	 * @param log the logger
	 */
	public SQLParser(ILogger log)
	{
		this.log = log;
	}
	/**
	 * builds the ANTLR parser from the sql expression
	 * @param sql the SQL query
	 * @return the parser object
	 * @throws ISQLException 
	 */
	public ISQLSelectParser buildParser(String sql) throws ISQLException
	{
		log.info("Parsing sql: "+sql);
		ISQLSelectLexer lex = new ISQLSelectLexer(new ANTLRStringStream(sql));
	    CommonTokenStream tokens = new CommonTokenStream(lex);
	    ISQLSelectParserWithErrors g = new ISQLSelectParserWithErrors(tokens);
	    try 
	    {
	    	g.statement();  
	    	List<RecognitionException> lstException = g.getExceptions();
	    	if(lstException.size() > 0)
	    	{
	    		String exceptionText = buildExceptionsText(lstException);
				throw new ISQLException("The sql query\n"+sql+"\nwas not parsed correctly. Error: "+exceptionText);
	    	}
	    } 
	    catch (RecognitionException e) 
	    {
	        throw new ISQLException("The sql query\n"+sql+"\nwas not parsed correctly. Error: "+e.getMessage(), e);
	    }
		return g;
	}

	/**
	 * Builds the text from a list of RecognitionException thrown by the parser
	 * @param lstException the list of exceptions returned by the parser
	 * @return a text including all the exceptions reported
	 */
	private String buildExceptionsText(
			List<RecognitionException> lstException)
	{
		StringBuilder builder = new StringBuilder(); 			
		for(RecognitionException ex : lstException)			
		{
			log.error("SQL parser exception: ", ex);
			String exceptionText = buildExceptionText(ex);
			builder.append(exceptionText).append('\n');
		}
		return builder.toString();
	}

	/**
	 * Builds the text from a RecognitionException thrown by the parser.
	 * The difference is that this is the exception THROWN by the parser, not returned.
	 * @param ex the exception
	 * @return the exception text
	 */
	private String buildExceptionText(RecognitionException ex)
	{
		String additionalInfo = null;
		if(ex instanceof MismatchedTokenException)
		{
			MismatchedTokenException mex = (MismatchedTokenException) ex;
			additionalInfo = "Expecting token "+ISQLSelectParser.tokenNames[ mex.expecting]+" got "+ISQLSelectParser.tokenNames[ mex.c];
		}
		String exceptionMessage = "At line: "+ex.line+" position: "+ex.charPositionInLine+": ";
		if(additionalInfo != null)
			exceptionMessage += additionalInfo+": ";
		return exceptionMessage;
	}

}
