/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */
package isql.builder;

import isql.Column;
import isql.ColumnKey;
import isql.ColumnValues;
import isql.IConstant;
import isql.IExpression;
import isql.ISQLException;
import isql.ISQLSelectParser;
import isql.constant.BooleanConstant;
import isql.constant.DateConstant;
import isql.constant.NumericConstant;
import isql.constant.TextConstant;
import isql.expression.ColumnExpressionSimple;
import isql.expression.ConstantExpression;
import isql.expression.DoubleExpression;
import isql.expression.UnaryExpression;
import isql.unary.IUnary;
import isql.unary.Minus;
import isql.unary.Not;

import java.sql.Timestamp;

import isql.help.scalar.IBinaryScalar;

/**
 * From the parsed expressions (in the select part or in the where part of
 * the SQL statement) extracts expression trees that can be used to calculate the result set 
 * @author SHZ Oct 11, 2007
 */
public abstract class ExpressionBuilder
{
	protected BinaryOperatorBuilder binaryOperatorBuilder = new BinaryOperatorBuilder();

	/**
	 * The original array values passed to the library for each column  
	 */
	protected ColumnValues values;
	/**
	 * Constructor
	 * @param values the columns built from the arrays matching the select items
	 */
	public ExpressionBuilder(ColumnValues values)
	{
		this.values = values;
	}
	
	/**
	 * The main method to extract the expression tree from an ANTLR parsed expression.	 
	 * @param expression the ANTLR expression
	 * @return the expression tree that will be used to execute the query
	 * @throws ISQLException
	 */
	public IExpression buildExpression(ISQLSelectParser.IExpression expression) throws ISQLException
	{
		return buildExpressionInternal(expression, new ExpressionStack());
	}
	
	/**
	 * Differentiate between simple expressions and complex expressions (operations between
	 * other expressions)
	 * @param expression the ANTLR expression
	 * @param parents the stack to know which expressions contain the expression we are converting
	 * @return the resulting expression tree
	 * @throws ISQLException
	 */
	protected IExpression buildExpressionInternal(ISQLSelectParser.IExpression expression, ExpressionStack parents) throws ISQLException
	{
		if(expression instanceof ISQLSelectParser.SimpleExpression)
			return buildSimpleExpression((ISQLSelectParser.SimpleExpression)expression, parents);
		else if(expression instanceof ISQLSelectParser.ComplexExpression)
			return buildComplexExpression((ISQLSelectParser.ComplexExpression)expression, parents);
		else
			throw new AssertionError("Expression type of"+expression.getClass()+" not recognized");
	}

	/**
	 * Builds the expression tree for a complex expression. A complex expression is an operation
	 * between two or more expressions.
	 * Extracts the operator and the two sub-expressions and acts recursively on them.
	 * @param expression the expression to convert
	 * @param parents the stack to know which expressions contain the expression we are converting 
	 * @return the result expression
	 * @throws ISQLException 
	 */
	protected  IExpression buildComplexExpression(
			ISQLSelectParser.ComplexExpression expression, ExpressionStack parents) throws ISQLException
	{	
		ExpressionStack newParents = parents.add(ExpressionBuilderTypes.BINARY);
		IExpression convA = buildExpressionInternal(expression.getA(), newParents);
		IExpression convB = buildExpressionInternal(expression.getB(), newParents);
		IBinaryScalar convOperator = binaryOperatorBuilder.buildBinaryOperator(expression.getOperator());
		return new DoubleExpression(convA, convB, convOperator);
	}
	

	/**
	 * extracts the expression tree from an unary expression (without operators or with an unary operator
	 * in front of it)
	 * @param expression the ANTLR expression
	 * @param parents the stack to know which expressions contain the expression we are converting
	 * @return the resulting expression tree
	 * @throws ISQLException 
	 */
	protected  IExpression buildSimpleExpression(ISQLSelectParser.SimpleExpression expression, ExpressionStack parents) throws ISQLException
	{
		ISQLSelectParser.SubExpression subExpression = expression.getSubExpression();				
		IExpression convExpression = buildSimpleExpressionWithoutUnary(subExpression, parents.add(ExpressionBuilderTypes.UNARY));
		String unaryOperator = subExpression.getUnaryOperator();
		if(unaryOperator != null)
		{
			IUnary convUnary = buildUnaryOperator(unaryOperator);
			return new UnaryExpression(convExpression, convUnary);
		}
		else
			return convExpression;
	}

	/**
	 * Converts the unary operator char to an internal constant
	 * @param unaryOperator the text defining the unary operator coming from ANTLR
	 * @return an enumerator defining the unary operator
	 * @throws ISQLException 
	 */
	protected IUnary buildUnaryOperator(String unaryOperator) throws ISQLException
	{
		if(unaryOperator.equals("-"))
			return new Minus();
		else if(unaryOperator.equals("~"))
			return new Not();
		else
			throw new ISQLException("Unary operator "+unaryOperator+" not recognized");
			
	}

	/**
	 * A simple expression can be a constant, an expression in parentheses or a table column.
	 * Extracts the expression tree for each case 
	 * @param subExpression the ANTLR expression
	 * @param parents the stack to know which expressions contain the expression we are converting
	 * @return the resulting expression tree
	 * @throws ISQLException
	 * @throws AssertionError
	 */
	protected IExpression buildSimpleExpressionWithoutUnary(
			ISQLSelectParser.SubExpression subExpression, ExpressionStack parents)
			throws ISQLException, AssertionError
	{
		if(subExpression instanceof ISQLSelectParser.ConstantSubExpression)
			return new ConstantExpression(
					buildConstantSubExpr((ISQLSelectParser.ConstantSubExpression)subExpression)
					);
		else if(subExpression instanceof ISQLSelectParser.ExpressionSubExpression)
			return buildExpressionInternal(
					((ISQLSelectParser.ExpressionSubExpression)subExpression).getExpression(), parents
					);
		else if(subExpression instanceof ISQLSelectParser.TableColumnSubExpression)
			return buildTableColumnSubExpr((ISQLSelectParser.TableColumnSubExpression)subExpression, parents);
		else
			throw new AssertionError("Sub expression "+subExpression.getClass()+" not recognized");
	}

	/**
	 * The expression is a simple table column. Just returns an object defining the column.
	 * @param subExpression the ANTLR expression
	 * @param parents the stack to know which expressions contain the expression we are converting
	 * @return the resulting expression tree
	 */
	private  IExpression buildTableColumnSubExpr(
			ISQLSelectParser.TableColumnSubExpression subExpression, ExpressionStack parents)
	{
		ISQLSelectParser.TableColumn tableColumn = subExpression.getTableColumn();
		ColumnKey key = buildTableColumn(tableColumn);
		return buildColumnExpressionFromKey(key, parents);
	}

	/**
	 * Returns the expression tree related to a table column
	 * @param key the column identifier (<table>.<column>)
	 * @param parents the stack to know which expressions contain the expression we are converting
	 * @return the resulting expression tree
	 */
	protected IExpression buildColumnExpressionFromKey(ColumnKey key, ExpressionStack parents)
	{
		return new ColumnExpressionSimple(new Column(key, values.get(key)));
	}

	/**
	 * Builds a table column object from the column identifier coming from ANTLR
	 * @param tableColumn the ANTLR column identifier
	 * @return the column identifier
	 */
	public static ColumnKey buildTableColumn(
			ISQLSelectParser.TableColumn tableColumn)
	{
		ColumnKey key = new ColumnKey(tableColumn.getTable(), tableColumn.getColumn());
		return key;
	}


	/**
	 * Builds the expression tree for a single constant
	 * @param subExpression the constant expression coming from ANTLR
	 * @return the constant (a text, number...) 
	 * @throws ISQLException 
	 * @throws AssertionError 
	 */
	protected  IConstant buildConstantSubExpr(ISQLSelectParser.ConstantSubExpression subExpression) throws AssertionError, ISQLException
	{
		ISQLSelectParser.IConstant constant = subExpression.getConstant();
		return buildConstant(constant);
		
	}

	/**
	 * Builds a constant object from an ANTLR parsed constant (t can be a numeric, text, date, boolean)
	 * @param constant the constant coming from ANTLR
	 * @return the constant (text, number...)
	 * @throws AssertionError
	 * @throws ISQLException
	 */
	public  IConstant buildConstant(ISQLSelectParser.IConstant constant)
			throws AssertionError, ISQLException
	{
		if(constant instanceof ISQLSelectParser.NumberConstant)
			return new NumericConstant(((ISQLSelectParser.NumberConstant)constant).getNumber());
		else if(constant instanceof ISQLSelectParser.StringConstant)
		{
			String text = ((ISQLSelectParser.StringConstant)constant).getText();
			String textWithoutQuotes = removeDelimiters(text, '\'');
			return new TextConstant(textWithoutQuotes);
		}
		else if(constant instanceof ISQLSelectParser.DateConstant)
		{
			String sDate = ((ISQLSelectParser.DateConstant)constant).getDate();
			String sDateWithoutDash = removeDelimiters(sDate, '#');
			try
			{
				Timestamp date = fromPartialString(sDateWithoutDash);
				return new DateConstant(date);
			}
			catch (IllegalArgumentException e)
			{
				throw new ISQLException("The date constant "+sDate+" does not contain a date");				
			}
		}
		else if(constant instanceof ISQLSelectParser.BooleanConstant)
		{
			boolean value = ((ISQLSelectParser.BooleanConstant)constant).getValue();
			return new BooleanConstant(value);
		}
		else
			throw new AssertionError("Unrecognized constant type "+constant.getClass());
	}

	/**
	 * Removes the delimiters of a text (quotes) or of a date (# characters)
	 * @param text a text surrounded by delimiters
	 * @param delimiter the delimiter character
	 * @return the text without delimiters 
	 * @throws ISQLException
	 */
	private String removeDelimiters(String text, char delimiter) throws ISQLException
	{
		int lastIndex = text.length() - 1;
		if(text.length() < 2 || text.charAt(0) != delimiter || text.charAt(lastIndex)  != delimiter)
			throw new ISQLException("The string constant "+text+" is not delimited by the \""+delimiter+"\" character");			
		String textWithoutQuotes = text.substring(1, lastIndex);
		return textWithoutQuotes;
	}
	
	/**
	 * extracts a date (Timestamp) from a date string
	 * @param partial the date string
	 * @return the extracted date
	 * @throws IllegalArgumentException
	 */
	public static Timestamp fromPartialString(String partial) throws IllegalArgumentException
	{
		String partialTrimmed = partial.trim();
		if(partialTrimmed.length() < 10)
			throw new IllegalArgumentException("The date ("+partial+")must at least contain year, month and day (YYYY-MM-DD)");
		return Timestamp.valueOf(partialTrimmed + "0000-00-00 00:00:00.000000000".substring(partialTrimmed.length()));					
	}

}
