/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */
package isql.builder;

import isql.Column;
import isql.ColumnKey;
import isql.ColumnValues;
import isql.IComparison;
import isql.IConstant;
import isql.IExpression;
import isql.ISQLException;
import isql.ISQLSelectParser;
import isql.expression.AggregateFunction;
import isql.expression.AggregateType;
import isql.expression.CaseExpression;
import isql.expression.ColumnExpressionByGroup;
import isql.expression.When;
import isql.expression.WhenThen;
import isql.groupby.Groups;

import java.util.ArrayList;
import java.util.List;


/**
 * The expression builder specifically for the select part of the SQL.
 * The difference from the where expression builder is that here you need to match the 
 * select expressions with the groups of the GROUP BY clause.
 * Also, CASE expressions are only valid in the select part of the query, not 
 * in the where clause
 * @author SHZ Oct 31, 2007
 */
public class SelectExpressionBuilder extends ExpressionBuilder
{
	/**
	 * The groups deriving from the GROUP BY clauses,
	 * which have to match the expressions in the select part of the query
	 */
	private Groups groups;
	
	public SelectExpressionBuilder(ColumnValues values, Groups groups)
	{
		super(values);
		this.groups = groups;
	}
	
	/**
	 * Builds a case expression tree from the ANTLR case expression 
	 * @param subExpression the ANTLR case expression
	 * @throws ISQLException the expression stack
	 */
	private CaseExpression buildCaseSubExpression(ISQLSelectParser.CaseSubExpression subExpression, ExpressionStack parents) throws ISQLException
	{
		ExpressionStack newParents = parents.add(ExpressionBuilderTypes.CASE);
		ISQLSelectParser.CaseFunction caseFunction = subExpression.getCase();
		ColumnKey testKey = buildTableColumn(caseFunction.getTest());
		IExpression convElse = buildExpressionInternal(caseFunction.getElse(), newParents);
		List<WhenThen> lstWhenThen = new ArrayList<WhenThen>();
		for(ISQLSelectParser.WhenThen whenThen : (List<ISQLSelectParser.WhenThen>)caseFunction.getWhenThens())
		{
			When convWhen = buildWhen(testKey, whenThen.getWhen());
			IExpression convThen = buildExpressionInternal(whenThen.getThen(), newParents);
			lstWhenThen.add(new WhenThen(convWhen, convThen));
		}
		Column testColumn = new Column(testKey, values.get(testKey));
		return new CaseExpression(testColumn, convElse, lstWhenThen);
	}

	/**
	 * Builds the when part of a CASE expression from the ANTLR when expression
	 * @param when the ANTLR when expression
	 * @return the when expression tree
	 * @throws ISQLException 
	 */
	private When buildWhen(ColumnKey testKeyCase, isql.ISQLSelectParser.When when) throws ISQLException
	{
		ColumnKey testKeyWhen = buildTableColumn(when.getTest());
		if(!testKeyCase.equals(testKeyWhen))
			throw new ISQLException("The case field "+testKeyCase+" is different than the when field "+testKeyWhen);
		IConstant compared = buildConstant(when.getCompared());
		IComparison operator = ComparisonBuilder.buildComparisonOperator(when.getOperator(), false);		
		return new When(operator, compared);
	}
	
	/**
	 * Builds an expression tree for an ANTLR expression that consists of a function with its parameters
	 * @param subExpression the ANTLR expression to convert
	 * @param parents the expression stack
	 * @return the function expression tree 
	 * @throws ISQLException 
	 */
	private  IExpression buildFunctionSubExpr(ISQLSelectParser.FunctionSubExpression subExpression, ExpressionStack parents) throws ISQLException
	{
		ISQLSelectParser.Function function = subExpression.getFunction();
		String functionId = function.getFunctionId();
		if(parents.contains(ExpressionBuilderTypes.FUNCTION))
			throw new ISQLException("Function "+functionId+": nested functions not allowed");		
		AggregateType aggregateType;
		try
		{
			aggregateType = AggregateType.valueOf(functionId.toUpperCase());
		}
		catch (RuntimeException e)
		{
			throw new ISQLException("Function name "+functionId+" not recognized");
		}
		if(function.getParameters().size() != 1)
			throw new ISQLException("The function "+functionId+" must have one parameter");
		ISQLSelectParser.IExpression parameter = (ISQLSelectParser.IExpression) function.getParameters().get(0);
		IExpression convParameter = super.buildExpressionInternal(parameter, parents.add(ExpressionBuilderTypes.FUNCTION));
		return new AggregateFunction(convParameter, groups, aggregateType);
	}

	/**
	 * builds an expression tree from an ANTLR simple expression.
	 * Filters the cases that the simple expression is a CASE, an expression composed by a function call and
	 * relies to the ExpressionBuilder.buildSimpleExpressionWithoutUnary for the other possibilities
	 * @param subExpression the ANTLR expression to convert
	 * @param parents the expression stack
	 * @return the expression  tree
	 */
	@Override
	protected IExpression buildSimpleExpressionWithoutUnary(
			ISQLSelectParser.SubExpression subExpression,
			ExpressionStack parents)
			throws ISQLException, AssertionError
	{
		if(subExpression instanceof ISQLSelectParser.CaseSubExpression)
			return buildCaseSubExpression((ISQLSelectParser.CaseSubExpression)subExpression, parents);
		else if(subExpression instanceof ISQLSelectParser.FunctionSubExpression)
			return buildFunctionSubExpr((ISQLSelectParser.FunctionSubExpression)subExpression, parents);
		else
			return super.buildSimpleExpressionWithoutUnary(subExpression, parents);
	}

	/**
	 * Builds an expression from ANTLR containing only a reference to a column
	 * to an expression tree.
	 * If the column is not a direct or indirect parameter of a function, it is connected with the related GROUP BY group
	 * (if there is a group by in the SQL query). 
	 */
	@Override
	protected IExpression buildColumnExpressionFromKey(ColumnKey key,
			ExpressionStack parents)
	{
		// if the column is a direct or indirect parameter of an aggregate function, the result
		// must contain all the rows of the where view.
		// if instead it is not aggregated, it must be a group by column and therefore
		// return only the summary values (one line for each group)		
		if(parents.contains(ExpressionBuilderTypes.FUNCTION) || groups == null)
			return super.buildColumnExpressionFromKey(key, parents);
		else
			return new ColumnExpressionByGroup(key, groups);			
	}
	

}

