/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>.
 *  
 * Released under the terms of the GNU LGPL v3.0.
*/
package isql.builder;

import isql.ColumnValues;
import isql.IComparison;
import isql.IConstant;
import isql.IExpression;
import isql.IPredicate;
import isql.ISQLException;
import isql.ISQLSelectParser;
import isql.predicate.AndPredicate;
import isql.predicate.ComparisonPredicate;
import isql.predicate.InPredicate;
import isql.predicate.LikePredicate;
import isql.predicate.OrPredicate;

import java.util.ArrayList;
import java.util.List;


/**
 * Builds the where clause.
 * It is composed by sub-clauses connected with the operators AND, OR
 * and negated with the operator NOT
 * The expressions are similar to the ones in the select part of the query.
 * but with small differences (no CASE expressions).
 * @author SHZ Oct 10, 2007
 */
public class WhereBuilder 
{
	/**
	 * The "where" clause contains some expression that needs to be evaluated
	 */
	private WhereExpressionBuilder expressionBuilder;
	/**
	 * the column/rows values that we apply the select query to
	 */
	private ColumnValues values;
	
	public WhereBuilder(ColumnValues values)
	{
		this.values = values;
		expressionBuilder = new WhereExpressionBuilder(values);
	}
	
	/**
	 * The method to call to evaluate a whole where clause
	 * @param condition the original where clause coming from ANTLR
	 * @return the built predicate tree resulting 
	 * @throws ISQLException
	 */
	public IPredicate build(ISQLSelectParser.ISearchCondition condition) throws ISQLException
	{
		return buildSearchCondition(condition, false);
	}
	
	/**
	 * Evaluates a search condition (a part of the where clause)
	 * @param condition the condition coming from ANTLR
	 * @param negated tells if this condition is negated by a NOT. 
	 * @return the condition tree
	 * @throws ISQLException
	 */
	private IPredicate buildSearchCondition(ISQLSelectParser.ISearchCondition condition, boolean negated) throws ISQLException
	{
		if(condition instanceof ISQLSelectParser.SimpleSearchCondition)
			return buildSimple((ISQLSelectParser.SimpleSearchCondition)condition, negated);
		else if(condition instanceof ISQLSelectParser.ComplexSearchCondition)
			return buildComplex((ISQLSelectParser.ComplexSearchCondition)condition, negated);
		else
			throw new AssertionError("Type of ISearchCondition "+condition.getClass()+" not recognized");
	}
			
	/**
	 * Evaluates a simple search condition (not "and" or "or")
	 * @param condition
	 * @param negated if the condition is negated twice (two NOTs before it) it is like it is not
	 * 		negated at all. This variable keeps track of how many times the expressions are negated
	 * 		so that we know if it is negated or not   
	 * @return the condition tree
	 * @throws ISQLException
	 */
	private IPredicate buildSimple(ISQLSelectParser.SimpleSearchCondition condition, boolean negated) throws ISQLException
	{
		ISQLSelectParser.SubSearchCondition sub = condition.getSubSearchCondition();
		if(sub.getNegated())
			negated = !negated;
		if(sub instanceof ISQLSelectParser.PredicateCondition)
			return buildPredicate((ISQLSelectParser.PredicateCondition)sub, negated);
		else if(sub instanceof ISQLSelectParser.RecursiveCondition)
			return buildSimpleRecursive((ISQLSelectParser.RecursiveCondition)sub, negated);
		else
			throw new AssertionError("Type of SubSearchCondition "+sub.getClass()+" not recognized");
	}
	
	/**
	 * Evaluates a search condition that contains another search condition.
	 * This is usually done because the original condition consist of the recursive condition
	 * with  a "not" in front of it
	 * @param sub the condition
	 * @param negated keeps track of the recursive negation
	 * @return
	 * @throws ISQLException 
	 */
	private IPredicate buildSimpleRecursive(ISQLSelectParser.RecursiveCondition sub,
			boolean negated) throws ISQLException
	{
		if(sub.getNegated())
			negated = !negated;
		return buildSearchCondition(sub.getSearchCondition(), negated);
	}

	/**
	 * A complex condition is expressed as e.g. and(or(a, d), or(b, c))
	 * So we need a recursive function to explore the whole condition 
	 * @param condition the ANTLR condition to parse
	 * @param operator the operator
	 * @param negated keeps track of the recursive negation
	 * @param lstOut
	 * @throws ISQLException
	 */
	private void buildComplexRecursive(ISQLSelectParser.ComplexSearchCondition condition, String operator, boolean negated, List<IPredicate> lstOut) throws ISQLException
	{
		ISQLSelectParser.ISearchCondition a = condition.getA();
		ISQLSelectParser.ISearchCondition b = condition.getB();
		buildComplexRecursiveSubCondition(condition, a, operator, negated, lstOut);
		buildComplexRecursiveSubCondition(condition, b, operator, negated, lstOut);		
	}

	/**
	 * The idea here is to push any "not" to the leafs of the tree resulting from the 
	 * where clause evaluation, and apply it there.
	 * Also, if since and(and(a,b),c) = and(a,b,c) and or(or(a,b),c) = or(a,b,c)
	 * we try to return a,b,c together in a list if we find these cases.  
	 * @param mainCondition the condition that is containing the current (subCondition)
	 * @param subCondition the current condition
	 * @param operator the operator of the complex condition (e.g and for and(..., ...)) that is the root of the expression 
	 * @param negated keeps track of the recursive negation
	 * @param lstOut the conditions 
	 * @throws ISQLException
	 */
	private void buildComplexRecursiveSubCondition(
			ISQLSelectParser.ComplexSearchCondition mainCondition,
			ISQLSelectParser.ISearchCondition subCondition, String operator, boolean negated,
			List<IPredicate> lstOut) throws ISQLException
	{
		if(subCondition instanceof ISQLSelectParser.ComplexSearchCondition)
		{
			ISQLSelectParser.ComplexSearchCondition complexSubCondition = (ISQLSelectParser.ComplexSearchCondition) subCondition;
			// if it is and(and(and(a,b),c),or(d,e)) we return a list [a,b,c,or(d,e)]
			if(mainCondition.getOperator().equalsIgnoreCase(operator))
				buildComplexRecursive(complexSubCondition, operator, negated, lstOut);
			else
			{
				// here there is shortcut. It is something like the or and(or(a,b),c).
				// we just have to build our predicate and return it.
				// lstOut.add(buildSimple((ISQLSelectParser.SimpleSearchCondition)subCondition, negated));
				IPredicate predicate = buildComplex(complexSubCondition, negated);
				lstOut.add(predicate);
			}
		}
		else 
		{
			lstOut.add(buildSearchCondition(subCondition, negated));
		}
	}
	
	/**
	 * Builds a complex condition (a binary condition with AND or OR operators) from the ANTLR expression
	 * @param condition the ANTLR expression
	 * @param negated keeps track of the recursive negation
	 * @return the expression tree
	 * @throws ISQLException
	 */
	private IPredicate buildComplex(ISQLSelectParser.ComplexSearchCondition condition, boolean negated) throws ISQLException
	{
		String operator = condition.getOperator().toLowerCase();
		List<IPredicate> lstCondition = new ArrayList<IPredicate>();
		buildComplexRecursive(condition, operator, negated, lstCondition);
		if(!operator.equals("and") && !operator.equals("or"))
			throw new AssertionError("Complex condition operator (should be and or or) not recognized");					
		boolean isAnd = operator.equalsIgnoreCase("and") && !negated; 
		if(isAnd)
			return new AndPredicate(lstCondition, values);
		else
			return new OrPredicate(lstCondition, values);
	}
	
	/**
	 * Builds a condition with the LIKE operator from the ANTLR expression
	 * It is in the form <toTest> LIKE <compared>
	 * @param predicate the LIKE expression coming from ANTLR
	 * @param negated keeps track of the recursive negation
	 * @return the Like predicate
	 * @throws ISQLException
	 */
	private IPredicate buildLike(ISQLSelectParser.Like predicate, boolean negated) throws ISQLException
	{
		IExpression toTest = expressionBuilder.buildExpression(predicate.getTest());
		return new LikePredicate(toTest, predicate.getCompared(), values, negated);
	}
	
	/**
	 * Builds an IN predicate from an ANTLR expression  
	 * @param predicate in the form of <toTest> IN (<constant1>...<constantN>) 
	 * @param negated keeps track of the recursive negation
	 * @return the in predicate
	 * @throws ISQLException
	 */
	private IPredicate buildIn(ISQLSelectParser.In predicate, boolean negated) throws ISQLException
	{
		List<ISQLSelectParser.IConstant> lstConstants = predicate.getConstants();
		ISQLSelectParser.IExpression test = predicate.getTest();
		IExpression convTest = expressionBuilder.buildExpression(test);
		IConstant[] arrConvConstant = new IConstant[lstConstants.size()];
		for(int i = 0; i < lstConstants.size(); i++)
		{
			ISQLSelectParser.IConstant constant = lstConstants.get(i);
			IConstant convConstant = expressionBuilder.buildConstant(constant);
			arrConvConstant[i] = convConstant;
		}
		InPredicate convPredicate = new InPredicate(convTest, arrConvConstant, values, negated);
		return convPredicate;
	}
	
	/**
	 * IN and LIKE are called NegatePredicates, because they can be preceded by a NOT and,
	 * unlike other predicates, cannot resolve the negation at this point (e.g. NOT(a = b) -> a <> b)
	 * but only when they are evaluated 
	 * @param predicate the IN or LIKE predicated from ANTLR
	 * @param negated keeps track of the recursive negation
	 * @return the IN or LIKE predicate
	 * @throws ISQLException
	 */
	private IPredicate buildNegatePredicate(ISQLSelectParser.NegatePredicate predicate, boolean negated) throws ISQLException
	{
		if(predicate.isNegated())
			negated = !negated;
		if(predicate instanceof ISQLSelectParser.Like)
			return buildLike((ISQLSelectParser.Like)predicate, negated);
		else if(predicate instanceof ISQLSelectParser.In)
			return buildIn((ISQLSelectParser.In)predicate, negated);
		else
			throw new AssertionError("Type of NegatePredicate "+predicate.getClass()+" not recognized");					
	}
	
	/**
	 * Builds a predicate from the ANTLR one
	 * @param condition the condition from ANTLR
	 * @param negated keeps track of the recursive negation
	 * @return the predicate
	 * @throws ISQLException
	 */
	private IPredicate buildPredicate(ISQLSelectParser.PredicateCondition condition, boolean negated) throws ISQLException
	{
		ISQLSelectParser.IPredicate predicate = condition.getPredicate();
		if(predicate instanceof ISQLSelectParser.Comparison)
			return buildComparison((ISQLSelectParser.Comparison)predicate, negated);
		else if(predicate instanceof ISQLSelectParser.NegatePredicate)
			return buildNegatePredicate((ISQLSelectParser.NegatePredicate)predicate, negated);
		else
			throw new AssertionError("Type of PredicateCondition "+condition.getClass()+" not recognized");
	}
	
	/**
	 * Builds a comparison condition (e.g. A = B, A <> B) from the ANTLR one 
	 * @param comparison the ANTLR comparison
	 * @param negated keeps track of the recursive negation
	 * @return the comparison condition
	 * @throws ISQLException
	 */
	private IPredicate buildComparison(ISQLSelectParser.Comparison comparison, boolean negated) throws ISQLException
	{
		IExpression a = expressionBuilder.buildExpression(comparison.getTest());
		IExpression b = expressionBuilder.buildExpression(comparison.getCompared());
		String operator = comparison.getOperator();
		IComparison comparisonOperator = ComparisonBuilder.buildComparisonOperator(operator, negated);
		ComparisonPredicate convComparison = new ComparisonPredicate(a, b, comparisonOperator);
		return convComparison;
		
	}

}
