/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */

package isql.expression;

import isql.ColumnKey;
import isql.IExpression;
import isql.ISQLException;
import isql.ITypedValue;
import isql.array.Array;
import isql.expression.aggregate.AggregateFactory;
import isql.expression.aggregate.IAggregate;
import isql.groupby.GroupKey;
import isql.groupby.Groups;
import isql.util.ArrayBuilder;

import java.util.List;
import java.util.Set;


/**
 * function to aggregate rows of the view (for example SUM, AVERAGE...).
 * It works in conjunction with the group by clause
 * @author SHZ Oct 1, 2007
 */
public class AggregateFunction implements ISimpleExpression
{
	/**
	 * The expression to aggregate
	 */
	private IExpression expression;
	/**
	 * The group by clause mapping
	 */
	private Groups groups;
	/**
	 * The function used to aggregate
	 */
	AggregateType aggregateType;
	
	public AggregateFunction(IExpression expression, Groups groups, AggregateType aggregateType)
	{
		this.expression = expression;
		this.groups = groups;
		this.aggregateType = aggregateType;
	}
	/* (non-Javadoc)
	 * @see isql.IExpression#getResult()
	 */
	public ITypedValue getResult() throws ISQLException
	{
		ITypedValue expressionResult = expression.getResult();
		if(!(expressionResult instanceof Array))
			throw new ISQLException("The expression to aggregate cannot be a constant");
		Array array = (Array) expressionResult;
		Object[] arrValue = array.getArray();
		List<GroupKey> lstGroupsInOriginalOrder = groups.getGroupKeysInOriginalOrder();
		Object[] arrRes = ArrayBuilder.buildArray(array.getType(), lstGroupsInOriginalOrder.size());
		int resIndex = 0;
		for(GroupKey groupKey : lstGroupsInOriginalOrder)
		{
			List<Integer> lstRowIndex = groups.getIndicesForGroupKey(groupKey);
			IAggregate aggregate = AggregateFactory.createAggregate(aggregateType);			
			for(Integer rowIndex : lstRowIndex)
			{
				Object value = arrValue[rowIndex.intValue()];
				aggregate.add(value);
			}
			arrRes[resIndex] = aggregate.getSum();
			resIndex++;
		}
		return new Array(arrRes, array.getType());
	}

	/* (non-Javadoc)
	 * @see isql.IExpression#getInvolvedColumns()
	 */
	public Set<ColumnKey> getInvolvedColumns()
	{
		return expression.getInvolvedColumns();
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		return "[AGGREGATE "+expression+" on "+groups+" with function "+aggregateType+"]";
	}

}
