/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */

package isql.expression;

import isql.Column;
import isql.ColumnKey;
import isql.Equalizer;
import isql.IExpression;
import isql.ISQLException;
import isql.ITypedValue;
import isql.Equalizer.IArrayFuture;
import isql.array.Array;
import isql.util.ArrayBuilder;

import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import isql.ColumnType;

/**
 * calculates the result of a CASE-WHEN-THEN expression
 * @author SHZ Oct 1, 2007
 */
public class CaseExpression implements ISimpleExpression
{
	/**
	 * The number of when-then clauses
	 */
	private int numWhenThen;
	/**
	 * The array of the when predicates
	 */
	private When[] arrWhen;
	/**
	 * The array (related to arrWhen) of the then expression results
	 */
	private Array[] arrThenArray;
	/**
	 * The else expression result
	 */		
	private Array elseArray;
	/**
	 * The type of each then and else expression, therefore the type of the case expression result
	 */
	private ColumnType type;
	/**
	 * The column that is tested against the then values
	 */
	private Column test;
	
	public CaseExpression(Column test, IExpression expElse, List<WhenThen> lstWhenThen) throws ISQLException, ISQLException
	{
		this.test = test;
		if(lstWhenThen.size() == 0)
			throw new ISQLException("The CASE expression does not have any WHEN-THEN parts");				
		numWhenThen = lstWhenThen.size();
		arrWhen = new When[numWhenThen];
		IArrayFuture[] arrThenFuture = new IArrayFuture[numWhenThen];
		Equalizer equalizer = new Equalizer();
		// if all the 'then' and the 'else' expressions are constants, the test columns is the only one that can give the correct size to all
		equalizer.add(test.getArray());
		ITypedValue elseVal = expElse.getResult();
		IArrayFuture elseFuture = equalizer.add(elseVal);
		type = test.getArray().getType();
		ColumnType elseType = elseVal.getType();
		if(type != elseType)
			throw new ISQLException("The case expression else expression must have the same values as the final column. Final column has type "+type+", else has type "+elseType);
			
		
		for(int i = 0; i < numWhenThen; i++)
		{
			WhenThen whenThen = lstWhenThen.get(i);
			When when = whenThen.getWhen();
			arrWhen[i] = when;
			IExpression then = whenThen.getThen();
			ITypedValue thenVal = then.getResult();
			if(thenVal.getType() != type)
				throw new ISQLException("The case expression then expression must have the same values as the final column. Final column has type  "+type+", then["+i+"] has type "+thenVal.getType());
			arrThenFuture[i] = equalizer.add(thenVal);
		}
		equalizer.equalize();
		arrThenArray = new Array[numWhenThen];
		for(int i = 0; i < numWhenThen; i++)
		{
			arrThenArray[i] = arrThenFuture[i].getArray();
		}
		elseArray = elseFuture.getArray();
	}
	
	
	/**
	 * @throws ISQLException
	 */
	/**
	 * Calculates the result of the case expression.
	 * The idea is that each then-when gives value fore some of the array's rows.
	 * We maintain the array usedIndex to know which rows of the result array got values, when-then after when-then.
	 * The rows that did not get values are covered by the else clause
	 * @return the result of the expression (can be an array or a constant)
	 * @throws ISQLException 
	 * @throws ISQLException 
	 */
	public ITypedValue getResult() throws ISQLException
	
	{
		// creates the usedIndex array, which contains  
		int size = elseArray.getSize();
		Object[] arrRes = ArrayBuilder.buildArray(test.getArray().getType(), size);
		for(int i = 0; i < numWhenThen; i++)
		{
			Object[] arrThen = arrThenArray[i].getArray();
			boolean[] useThis = arrWhen[i].apply(test);				
			for(int j = 0; j< size; j++)					
				if(arrRes[j] == null && useThis[j])
					arrRes[j] = arrThen[j]; 
		}
		Object[] arrElse = elseArray.getArray();
		for(int j = 0; j< size; j++)					
			if(arrRes[j] == null)
				arrRes[j] = arrElse[j]; 
		return new Array(arrRes, type);
	}


	/* (non-Javadoc)
	 * @see isql.IExpression#getInvolvedColumns()
	 */
	public Set<ColumnKey> getInvolvedColumns()
	{
		Set<ColumnKey> setColumnKey = new TreeSet<ColumnKey>();
		setColumnKey.add(test.getKey());
		return setColumnKey;
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		StringBuilder b = new StringBuilder();
		b.append("[CASE ");
		b.append("test : ").append(test);
		for(int i = 0; i < numWhenThen; i++)
		{
			When when = arrWhen[i];
			Array thenArray = arrThenArray[i];
			b.append(" When: ").append(when).append(" Then: ").append(thenArray);
		}
		b.append("Else: ").append(elseArray);
		b.append(']');
		return b.toString();
	}

	
}
