/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */

package isql.expression;

import isql.ColumnKey;
import isql.Equalizer;
import isql.IExpression;
import isql.ISQLException;
import isql.ITypedValue;
import isql.Equalizer.IArrayFuture;

import java.sql.Timestamp;
import java.util.Set;
import java.util.TreeSet;

import isql.array.Array;
import isql.help.scalar.IBinaryScalar;
import isql.help.scalar.bool.IBinaryBooleanScalar;
import isql.help.scalar.date.IBinaryDateScalar;
import isql.help.scalar.numeric.IBinaryNumericScalar;
import isql.help.scalar.text.IBinaryTextScalar;
import isql.ColumnType;

/**
 * The combination of two other expression through a
 * binary operator (e.g. a+b)
 * @author SHZ Sep 28, 2007
 */
public class DoubleExpression implements IExpression
{
	/**
	 * The first operand (it is an expression itself) 
	 */
	private IExpression a;
	/**
	 * the second operand (it is an expression itself) 
	 */
	private IExpression b;
	/**
	 * The operator (e.g. plus)
	 */
	private IBinaryScalar operator;
	
	public DoubleExpression(IExpression a, IExpression b,
			IBinaryScalar operator)
	{
		this.a = a;
		this.b = b;
		this.operator = operator;
	}

	public ITypedValue getResult() throws ISQLException
	{
		ITypedValue resA = a.getResult();
		ITypedValue resB = b.getResult();
		ColumnType typA = resA.getType();
		ColumnType typB = resB.getType();
		if(typA != typB)
			throw new ISQLException("Comparison: Different types "+typA+"<>"+typB);
		Equalizer equalizer = new Equalizer();
		IArrayFuture futureA = equalizer.add(resA);
		IArrayFuture futureB = equalizer.add(resB);
		try
		{
			equalizer.equalize();
			Array eqA = futureA.getArray();
			Array eqB = futureB.getArray();
			int sizeA = eqA.getSize();
			return calculate(typA, eqA, eqB, sizeA);
		}
		catch (ISQLException ex)
		{
			throw new ISQLException("Comparing the expressions "+a+ " and "+b+" got", ex);
		}
	}

	/**
	 * @param type
	 * @param eqA
	 * @param eqB
	 * @param size
	 * @return
	 * @throws ComparisonException
	 * @throws AssertionError
	 */
	private Array calculate(ColumnType type, Array eqA, Array eqB, int size)
			throws ISQLException, AssertionError
	{
		Object[] arrA = eqA.getArray(); 
		Object[] arrB = eqB.getArray();
		Object[] res;
		switch(type)
		{
			case NUMERIC:
			{
				res = new Double[arrA.length];
				for(int i = 0; i < size; i++)
					 res[i] = Double.valueOf( 
						 ((IBinaryNumericScalar)operator).apply(
							 ((Double)arrA[i]).doubleValue(),
							 ((Double)arrB[i]).doubleValue()
							 ));
				break; 
			}
			case TEXT:
			{
				res = new String[arrA.length];
				for(int i = 0; i < size; i++)
					 res[i] =  
						 ((IBinaryTextScalar)operator).apply(
							 (String)arrA[i],
							 (String)arrB[i]
							 );
				break;
			}
			case BOOLEAN:
			{
				res = new Boolean[arrA.length];				
				for(int i = 0; i < size; i++)
					 res[i] = Boolean.valueOf( 
						 ((IBinaryBooleanScalar)operator).apply(
							 ((Boolean)arrA[i]).booleanValue(),
							 ((Boolean)arrB[i]).booleanValue()
							 ));
				break;
			}
			case DATE:
			{
				res = new Timestamp[arrA.length];
				for(int i = 0; i < size; i++)
					 res[i] =  
						 ((IBinaryDateScalar)operator).apply(
							 (Timestamp)arrA[i],
							 (Timestamp)arrB[i]
							 );
				break;
			}
			default:
				throw new AssertionError("DoubleExpression : type not recognized "+type);
		}
		return new Array(res, type);
	}

	/* (non-Javadoc)
	 * @see isql.IExpression#getInvolvedColumns()
	 */
	public Set<ColumnKey> getInvolvedColumns()
	{
		Set<ColumnKey> setColumnKey = new TreeSet<ColumnKey>();
		setColumnKey.addAll(a.getInvolvedColumns());
		setColumnKey.addAll(b.getInvolvedColumns());
		return setColumnKey;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		StringBuilder b = new StringBuilder();
		b.append("[DOUBLEEXPRESSION ").append(a).append(' ').append(operator).append(' ').append(b).append(']');
		return b.toString();
	}
	
	

	
}
