package isql.groupby;

import java.util.Arrays;


/**
 * Used as key for the groups in the GROUP BY clause.
 * The GROUP BY clause divides the rows in sets; in each set the GROUP BY columns
 * have the same values for all rows, the other columns have different values.
 * The keys contain the invariant GROUP BY columns values. 
 * Assumes that the columns are always in the same positions in the array
 * @author SHZ Oct 29, 2007
 */
public class GroupKey implements Comparable<GroupKey>
{
	/**
	 * The values of the columns in the GROUP BY clause, which are invariant
	 * for a set of rows (the set of rows defined by the group)
	 */
	private Object[] arrColumnValue;
	
	public GroupKey(Object[] arrColumnValue)
	{
		this.arrColumnValue = arrColumnValue;
	}

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	public int compareTo(GroupKey o)
	{
		if(arrColumnValue.length != o.arrColumnValue.length)
			throw new AssertionError("The group keys "+this+" and "+o+" have a different number of columns");
		for(int i = 0; i < arrColumnValue.length; i++)
		{
			Comparable value = (Comparable) arrColumnValue[i];
			Comparable oValue = (Comparable) o.arrColumnValue[i];
			int compareRes = value.compareTo(oValue);
			if(compareRes != 0)
				return compareRes;
		}
		return 0;
	}

	/**
	 * The invariant columns values for the set identified by this key
	 * @return the arrColumnValue
	 */
	public Object[] getArrColumnValue()
	{
		return arrColumnValue;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(arrColumnValue);
		return result;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj)
	{
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		GroupKey other = (GroupKey) obj;
		if (!Arrays.equals(arrColumnValue, other.arrColumnValue))
			return false;
		return true;
	}
}