/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */
package isql.groupby;

import isql.ColumnKey;
import isql.ISQLException;
import isql.array.Array;
import isql.util.ArrayBuilder;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import isql.ColumnType;
import isql.util.MapOfListHelp;
import isql.util.StringJoin;

/**
 * Contains the groups coming from the "group by" clause.
 * The groups are represented as a map that, for each group, maps the values that are constant
 * in the group to the indices of the view lines that are in that group.  
 * @author SHZ Oct 29, 2007
 */
public class Groups
{
	/**
	 * maps a "group by" key (column values that are always the same for a group) with the indices
	 * of the view in which they are valid 
	 */
	private Map<GroupKey, List<Integer>> mapRowIndicesByGroup = new TreeMap<GroupKey, List<Integer>>();
	/**
	 * Contains the group by key, ordered by the first rows that was part of them. Used
	 * to maintain the (possible) order by order
	 */
	private Map<Integer, GroupKey> mapGroupKeyByFirstIndex = new TreeMap<Integer, GroupKey>();
	/**
	 * The column keys of the columns that define the "group by", 
	 * i.e. the keys of the columns that have constant values in the groups.
	 * They are in the same order as their values in GroupKey
	 */
	private ColumnKeyAndType[] arrGroupByColumnDef;
	
	public Groups(ColumnKeyAndType[] arrColumnDef)
	{
		this.arrGroupByColumnDef = arrColumnDef;
	}
	/**
	 * The given row index has to be part of the given group
	 * @param key
	 * @param rowIndex
	 */
	public void add(GroupKey key, int rowIndex)
	{
		if(!mapRowIndicesByGroup.containsKey(key))
			mapGroupKeyByFirstIndex.put(rowIndex, key);
		MapOfListHelp.put(mapRowIndicesByGroup, key, Integer.valueOf(rowIndex));
	}
	
	/**
	 * Tests if a column is a group by column, i.e. it was part of the 
	 * group by clause
	 * @param toTest the identifier of the column
	 * @return true if it is a group by column, false otherwise
	 */
	public boolean isGroupByColumnKey(ColumnKey toTest)
	{
		for(ColumnKeyAndType groupByColumnDef : arrGroupByColumnDef)
			if(toTest.equals(groupByColumnDef.getKey()))
				return true;
		return false;
	}
	
	private int getIndexOfColumnKey(ColumnKey key) throws ISQLException
	{
		for(int i = 0; i < arrGroupByColumnDef.length; i++)
		{
			ColumnKeyAndType groupDef = arrGroupByColumnDef[i];
			if(key.equals(groupDef.getKey()))
				return i;
		}
		throw new ISQLException("Key "+key+" is not a group by column");
	}
	/**
	 * Returns an array with the rows index that are contained in the group,
	 * in the order they was found in the original set of columns
	 * @param key the column identifier
	 * @return the array containing for each group the value that the column
	 * 	has in that group
	 * @throws ISQLException
	 */
	public Array getGroupByColumnValues(ColumnKey key) throws ISQLException
	{
		int columnIndex = getIndexOfColumnKey(key);
		ColumnType type = arrGroupByColumnDef[columnIndex].getType();
		Object[] arrColumnValue = ArrayBuilder.buildArray(type, mapGroupKeyByFirstIndex.size());
		int i = 0;
		for(GroupKey groupKey : mapGroupKeyByFirstIndex.values())
		{			
			Object value = groupKey.getArrColumnValue()[columnIndex];
			arrColumnValue[i] = value;
			i++;
		}
		return new Array(arrColumnValue, type);
	}
	
	public List<Integer> getIndicesForGroupKey(GroupKey key)
	{
		return mapRowIndicesByGroup.get(key);
	}
	
	public List<GroupKey> getGroupKeysInOriginalOrder()
	{
		return new ArrayList<GroupKey>(mapGroupKeyByFirstIndex.values());
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		StringBuilder b = new StringBuilder();
		b.append("[GROUPS ");
		b.append(" IndicesByGroup: ");
		for(Map.Entry<GroupKey, List<Integer>> entry : mapRowIndicesByGroup.entrySet())
		{
			GroupKey key = entry.getKey();
			List<Integer> lstIndex = entry.getValue();
			b.append(" Group: ").append(key.toString());
			b.append(" Indices: ").append(StringJoin.join(lstIndex, ','));
		}
		b.append(" Group by column def: ");
		for(ColumnKeyAndType keyAndType : arrGroupByColumnDef)
		{
			b.append(' ').append(keyAndType.toString());
		}
		b.append(']');
		return b.toString();
	}
	

}


