/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */

package isql.predicate;

import isql.ColumnType;
import isql.IComparison;
import isql.IConstant;
import isql.IExpression;
import isql.IPredicate;
import isql.ISQLException;
import isql.ITypedValue;
import isql.array.Array;

/**
 * It builds the view for a single comparison in the WHERE clause
 * (a = b, a <> b...)
 * @author SHZ Sep 27, 2007
 */
public class ComparisonPredicate implements IPredicate
{
	/**
	 * To compare with b
	 */
	private IExpression a;
	/**
	 * To compare with a
	 */
	private IExpression b;
	/**
	 * The comparison operator to apply, in form of a function
	 */
	private IComparison comparison;
	
	public ComparisonPredicate(IExpression a, IExpression b, IComparison comparison)
	{
		this.a = a;
		this.b = b;
		this.comparison = comparison;
	}
	
	/**
	 * Builds the view resulting from the comparison of the two expressions
	 * @return the view 
	 */
	public WhereView apply() throws ISQLException
	{
		ITypedValue resA = a.getResult();		
		ITypedValue resB = b.getResult();
		ColumnType typA = resA.getType();
		ColumnType typB = resB.getType();
		if(typA != typB)
			throw new ISQLException("Comparison: Different types "+typA+"<>"+typB);
		Object[] arrA = toArray(resA);
		Object[] arrB = toArray(resB);
		WhereView viewA = WhereView.fromExpression(a);
		WhereView viewB = WhereView.fromExpression(b);
		fillViewMaps(arrA, arrB, viewA, viewB, comparison);
		WhereView viewRet = WhereView.none(new String[] {});
		viewRet.addAll(viewA);
		viewRet.addAll(viewB);
		return viewRet;
	}

	/**
	 * Fills the views of A and B with the indices of the columns for the rows
	 * for which the comparison has success
	 * @param arrA the values of column A 
	 * @param arrB the values of column B
	 * @param viewA the view related to column A, that gets filled with the indices
	 * @param viewB the view related to column B, that gets filled with the indices
	 * @param comparison the comparison operator
	 * @throws ISQLException
	 */
	public static void fillViewMaps(Object[] arrA, Object[] arrB,
			WhereView viewA, WhereView viewB, 
			IComparison comparison) throws ISQLException 
	{
		for(int i = 0 ; i < arrA.length; i++)
			for(int j = 0; j < arrB.length; j++)
			{				
				if(comparison.compare(arrA[i], arrB[j]))
				{
					viewA.addRowIndex(i);
					viewB.addRowIndex(j);
				}
			}
	}


	/**
	 * Convert the content of a column into an array, whatever it is 
	 * @param res the content of a column (can be an array or a constant)
	 * @return the final array 
	 */
	public static Object[] toArray(ITypedValue res)
	{
		Object[] arrA;
		if(res instanceof Array)
			arrA = ((Array)res).getArray();
		else
			arrA = new Object[] { ((IConstant)res).getValueAsObject() };
		return arrA;
	}
	
	@Override
	public String toString()
	{
		return a+" "+comparison+" "+b;
	}

}
