/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */

package isql.predicate;
import isql.ColumnValues;
import isql.IPredicate;
import isql.ISQLException;
import isql.util.MapOfListHelp;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Returns a view that is the result of AND-ing (or OR-ing) all the views (table rows in relation) that 
 * return from the input conditions.
 * E.g.
 * WHERE T1.C1 = T2.C2
 * AND T2.C3 = T3.C4
 * 
 * the two predicates have the table T2 in common.
 * In this case the result contains all the view rows of the first predicate
 * that match the view rows of the second predicate that contain the same T2 table rows. 
 * @author SHZ Oct 1, 2007
 */
public class DoublePredicate
{
	
	/**
	 * Add row to the result view only at a certain condition 
	 * @author SHZ Oct 18, 2007
	 */
	public static interface IAddRow
	{
		/**
		 * Adds the row index to the result view 
		 * @param viewA the first view to merge
		 * @param viewB the second view to merge
		 * @param mergeTable the table in common between the two views that is used to merge them
		 * @param viewRes the result view
		 * @param viewRowA 
		 * @param lstViewRowBForMergeTableRowA the
		 */
		public void addRow(WhereView viewA, WhereView viewB, String mergeTable,
				WhereView viewRes, int viewRowA,
				List<Integer> lstViewRowBForMergeTableRowA);
		
	}
	
	/**
	 * Add row to the result view only if both the views have, for that row, the same index in the merge table
	 * @author SHZ Oct 18, 2007
	 */
	public static class AddRowOnlyIfInBoth implements IAddRow
	{
		public void addRow(WhereView viewA, WhereView viewB, String mergeTable,
				WhereView viewRes, int viewRowA,
				List<Integer> lstViewRowBForMergeTableRowA)
		{
			if(lstViewRowBForMergeTableRowA != null)
			{
				for(Integer viewRowB : lstViewRowBForMergeTableRowA)
					setRowInRes(viewRes, viewA, viewB, viewRowA, viewRowB.intValue(), mergeTable);
			}
		}		
	}
	
	/**
	 * Add row to the result view only if only one of the views have, for that row, a specific index in the merge table
	 * In this case the other table will have dummy values (-1) for the column indices in that row
	 * @author SHZ Oct 18, 2007
	 *
	 */
	public static class AddRowOnlyIfInJustInOne implements IAddRow
	{
		public void addRow(WhereView viewA, WhereView viewB, String mergeTable,
				WhereView viewRes, int viewRowA,
				List<Integer> lstViewRowBForMergeTableRowA)
		{
			if(lstViewRowBForMergeTableRowA == null)
			{
				setRowInRes(viewRes, viewA, viewB, viewRowA, WhereView.NO_ROW, mergeTable);
			}
		}		
	}

	/**
	 * Add always a row. The result is the one of AddRowOnlyIfInBoth and AddRowOnlyIfInJustInOne combined
	 * @author SHZ Oct 18, 2007
	 */
	public static class AddRowAnyway implements IAddRow
	{
		public void addRow(WhereView viewA, WhereView viewB, String mergeTable,
				WhereView viewRes, int viewRowA,
				List<Integer> lstViewRowBForMergeTableRowA)
		{
			if(lstViewRowBForMergeTableRowA != null)
			{
				for(Integer viewRowB : lstViewRowBForMergeTableRowA)
					setRowInRes(viewRes, viewA, viewB, viewRowA, viewRowB.intValue(), mergeTable);
			}
			else
				setRowInRes(viewRes, viewA, viewB, viewRowA, WhereView.NO_ROW, mergeTable);
		}		
	}
	
	
	/**
	 * The list of the predicates to "connect" with the AND/OR operator
	 */
	private List<IPredicate> lstPredicate;
	
	private IMergeRows mergeRows;
	
	private ColumnValues values;
	/* (non-Javadoc)
	 * @see isql.IPredicate#apply()
	 */
	public DoublePredicate(List<IPredicate> lstPredicate, IMergeRows mergeRows, ColumnValues values)
	{
		this.lstPredicate = lstPredicate;
		this.mergeRows = mergeRows;
		this.values = values;
	}
	
	/**
	 * applies the AND/OR to the predicates. 
	 * The final result is a "view" that is the result of putting in relation all predicates,
	 * two at a time, by a common table that they contain.
	 * For example:
	 * 
	 * WHERE T1.C1 = T2.C1 (predicate A)
	 * AND T2.C2 = T3.C1 (predicate B)
	 * 
	 * The two predicates contain the same table T2, which is used by the AND to create the view.
	 * The idea is to:
	 * - predicate A returns a view, containing all the rows of T1 and T2 that match the expression; we call it viewA
	 * - predicate B returns viewB.
	 * - put the two views in relation, connecting them by the common table T2. Which means connecting
	 *   the rows of viewA and viewB by the common T2 rows, building in that way the final view.
	 */
	public WhereView apply() throws ISQLException
	{
		if(lstPredicate.isEmpty())
			return WhereView.fromColumns(values);
		List<WhereView> lstOut = new ArrayList<WhereView>();
		for(IPredicate predicate : lstPredicate)
		{
			lstOut.add(predicate.apply());
		}					
		while(lstOut.size() > 1)
		{
			int last = lstOut.size() -1;
			WhereView toFind = lstOut.get(last);		
			lstOut.remove(last);
			mergePredicateWithCommonTables(toFind, lstOut, mergeRows);
		}
		// the only element remained after the merge		
		WhereView res = lstOut.get(0);
		return res;
	}
	
	/**
	 * Merge two views in one single view, by the common table
	 * @param viewA the first view to merge 
	 * @param viewB the second view to merge
	 * @param mergeTable the table that is common between the two views and is used to merge them
	 * @param viewRes the resulting view (the row indexes are added to it)
	 * @param addRow tells when the rows are added to the result view
	 */
	private static void mergeRowsOfBInA(WhereView viewA, WhereView viewB, String mergeTable, WhereView viewRes, IAddRow addRow)
	{
		Map<Integer, List<Integer>> mapViewRowByTableRowB = buildMapViewRowByTableRow(
				viewB, mergeTable);
		List<Integer> lstOfTableA = viewA.getIndicesInTable(mergeTable);
		// copies the view rows of A and B to the final map
		// There can be more than one viewRowA for each tableRowA
		for(int viewRowA = 0; viewRowA != lstOfTableA.size(); viewRowA++)  
		{
			Integer tableRowA = lstOfTableA.get(viewRowA);
			List<Integer> lstViewRowBForTableRowA = mapViewRowByTableRowB.get(tableRowA);
			addRow.addRow(viewA, viewB, mergeTable, viewRes, viewRowA,
					lstViewRowBForTableRowA);
		}
	}


	/**
	 * Builds the result view, initializing the view row lists for each table involved
	 * The result view has only tables, not indices
	 * @param viewA the first view to merge
	 * @param viewB the second view to merge
	 * @return the result view
	 */
	private WhereView buildResultView(WhereView viewA, WhereView viewB)
	{
		WhereView viewRes = WhereView.none(viewA.getTableNames());
		viewRes.addEmptyTables(viewB.getTableNames());
		return viewRes;
	}

	/**
	 * build a map to find the view rows that contain the rows of "mergeTable"
	 * @param view the view to check
	 * @param mergeTable the table in common with another view
	 * @return a map [merge table row] -> [ view rows that match the merge table row] 
	 */
	private static Map<Integer, List<Integer>> buildMapViewRowByTableRow(
			WhereView view, String mergeTable)
	{
		Map<Integer, List<Integer>> mapViewRowByTableRow = new HashMap<Integer, List<Integer>>();
		List<Integer> lstOfMergeTable = view.getIndicesInTable(mergeTable);
		for(int viewRow = 0; viewRow != lstOfMergeTable.size(); viewRow++)
		{
			Integer mergeTableRow = lstOfMergeTable.get(viewRow);
			MapOfListHelp.put(mapViewRowByTableRow, mergeTableRow, Integer.valueOf(viewRow));
		}
		return mapViewRowByTableRow;
	}
	/**
	 * Puts together in res one single row of viewA and one single row of viewB
	 * @param res the final view
	 * @param viewA the view to copy 
	 * @param viewB the other view to copy
	 * @param viewRowA the index of the row of viewA to copy
	 * @param viewRowB the index of the row of viewB to copy
	 * @param mergeTable the table that is in common between viewA and viewB
	 */
	private static void setRowInRes(WhereView res, WhereView viewA, WhereView viewB, int viewRowA, int viewRowB, String mergeTable)
	{
		for(String tableA: viewA.getTableNames())
		{
			List<Integer> lstForTableNameA = viewA.getIndicesInTable(tableA);
			List<Integer> lstForTableNameRes = res.getIndicesInTable(tableA);
			lstForTableNameRes.add(lstForTableNameA.get(viewRowA));
		}
		for(String tableB : viewB.getTableNames())
		{
			if(!tableB.equals(mergeTable))
			{
				List<Integer> lstForTableNameB = viewB.getIndicesInTable(tableB);
				List<Integer> lstForTableNameRes = res.getIndicesInTable(tableB);
				lstForTableNameRes.add(lstForTableNameB.get(viewRowA));
			}
		}						
	}

	
	/**
	 * Checks which predicate has a table in common with toFind and merges them in one view
	 * @param toFind the predicate to check 
	 * @param lstIn the list of predicates to look in
	 * @throws ISQLException
	 */
	private void mergePredicateWithCommonTables(WhereView toFind, List<WhereView> lstIn, IMergeRows mergeRows) throws ISQLException
	{
		boolean foundSomethingtoMerge = false;
		for(int i = 0; i < lstIn.size(); i++)
		{
			WhereView in = lstIn.get(i);
			for(String tableName : toFind.getTableNames())
			{
				if(in.hasTable(tableName))
				{
					WhereView viewRes = buildResultView(in, toFind);
					mergeRows.mergeRows(toFind, in, tableName, viewRes);
					lstIn.set(i, viewRes);
					foundSomethingtoMerge = true;
				}
			}				
		}
		// did not find anything. Very strange that there is a predicate that does not have a "friend" with one common table
		if(!foundSomethingtoMerge)
			throw new ISQLException("Predicate without friend");
	}

	/**
	 * In case of an AND between two conditions, the views have to be merged
	 * so that only the row indices that are in both views are stored in the result
	 * view
	 * @author SHZ Apr 16, 2008
	 */
	public static class MergeRowsAnd implements IMergeRows
	{	
		/* (non-Javadoc)
		 * @see isql.predicate.IMergeRows#mergeRows(isql.predicate.WhereView, isql.predicate.WhereView, java.lang.String, isql.predicate.WhereView)
		 */
		public void mergeRows(WhereView toFind, WhereView in, String tableName,
				WhereView viewRes)
		{
			mergeRowsOfBInA(in, toFind, tableName, viewRes, new AddRowOnlyIfInBoth());
		}
	}
	
	/**
	 * In case of an AND between two conditions, the views have to be merged
	 * so that the result view contains the union of the rows indices of the views.
	 * This is accomplished:
	 * 1) copying to the result view all the indices of the first view
	 * 2) copying to the result view the indices of the second view that are not in the first view
	 * @author SHZ Apr 16, 2008
	 */
	public static class MergeRowsOr implements IMergeRows
	{	
		public void mergeRows(WhereView toFind, WhereView in, String tableName,
				WhereView viewRes)
		{
			mergeRowsOfBInA(in, toFind, tableName, viewRes, new AddRowAnyway());
			mergeRowsOfBInA(toFind, in, tableName, viewRes, new AddRowOnlyIfInJustInOne());
		}
	}
	
	@Override
	public String toString()
	{
		// [a,b,c..]
		return lstPredicate.toString();
	}
	
}
