/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */

package isql.predicate;

import java.util.Arrays;

import isql.ColumnValues;
import isql.IConstant;
import isql.IExpression;
import isql.IPredicate;
import isql.ISQLException;
import isql.ITypedValue;
import isql.ColumnType;
import isql.array.Array;
import isql.util.BooleanHelp;

/**
 * In clause. Compares an expression (toTest) with the constants in a sequence (seqIn)
 * @author SHZ Sep 28, 2007
 */
public class InPredicate implements IPredicate
{
	/**
	 * The expression to test against the list of constant
	 */
	private IExpression toTest;
	/**
	 * The list of "in" constants
	 */
	private IConstant[] seqIn;
	/**
	 * If there is a not ("not in") before the in, applies the negation
	 */
	private boolean negated;
	/**
	 * The result of the "join" of the where condition
	 */
	private ColumnValues values;
	
	public InPredicate(IExpression toTest, IConstant[] seqIn, ColumnValues values, boolean negated)
	{
		this.toTest = toTest;
		this.seqIn = seqIn;
		this.negated = negated;
		this.values = values;
	}

	/**
	 * Builds the view resulting from the in predicate
	 * @return the result view
	 */
	public WhereView apply() throws ISQLException
	{
		ITypedValue valToTest = toTest.getResult();
		ColumnType type = valToTest.getType();
		for(IConstant in : seqIn)
			if(in.getType() != type)
				throw new ISQLException("The expression "+in+" has not the same type as the test expression in the IN clause");
		if(valToTest instanceof Array)
		{
			WhereView ret = WhereView.fromExpression(toTest);			
			Object[] arrToTest = ((Array)valToTest).getArray();
			for(int i = 0; i < arrToTest.length; i++)
			{
				if(BooleanHelp.xor(atLeastOne(arrToTest[i], seqIn), !negated))
					ret.addRowIndex(i);
			}
			return ret;			
		}
		else if(valToTest instanceof IConstant)
		{
			IConstant constToTest = (IConstant) valToTest;
			String[] tables = WhereView.getTablesFromColumnKeys(toTest.getInvolvedColumns());
			if(BooleanHelp.xor(atLeastOne(constToTest.getValueAsObject(), seqIn), !negated))
				return WhereView.fromColumns(values);
			else
				return WhereView.none(tables);
		}
		else
			throw new AssertionError("ITypedValue not recognized "+valToTest.getClass());
	}

	

	/**
	 * It is to check if a value matches at least one constant in the sequence after the IN
	 * clause
	 * @param val the value to check 
	 * @param seqIn the IN sequence
	 * @return true if it matches at least one, false otherwise
	 */
	private static boolean atLeastOne(Object val, IConstant[] seqIn)
	{
		for(IConstant in : seqIn)
		{
			Object inValue = in.getValueAsObject(); 
			if(inValue.equals(val))
				return true;
		}
		return false;
	}
	
	@Override
	public String toString()
	{
		String s = toTest.toString()+" IN "+Arrays.toString(seqIn);
		if(negated)
			return "NOT "+s;
		else
			return s;
	}
	
}
