/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */
package isql.predicate;

import isql.ColumnValues;
import isql.IConstant;
import isql.IExpression;
import isql.IPredicate;
import isql.ISQLException;
import isql.ITypedValue;
import isql.array.Array;
import isql.constant.TextConstant;

import java.util.regex.Pattern;

import isql.ColumnType;
import isql.util.BooleanHelp;

/**
 * exp LIKE <text>
 * exp can be a column or an expression that involves table columns
 * or a constant (which does actually not make so much sense)
 * @author SHZ Oct 11, 2007
 */
public class LikePredicate implements IPredicate
{
	/**
	 * The expression to test against the pattern
	 */
	private IExpression test;
	/**
	 * The LIKE pattern
	 */	
	private String like;
	/**
	 * Just to build the view containing everything if the test expression
	 * is a constant
	 */
	private ColumnValues values;
	
	/**
	 * If this is true, the predicate must be intended as "not like"
	 */
	private boolean negated;
	/**
	 * @param test
	 * @param like
	 */
	public LikePredicate(IExpression test, String like, ColumnValues values, boolean negated)
	{
		this.test = test;
		this.like = like;
		this.values = values;
		this.negated = negated;
	}
	/* (non-Javadoc)
	 * @see isql.IPredicate#apply()
	 */
	public WhereView apply() throws ISQLException
	{
		ITypedValue valTest = test.getResult();
		if(valTest.getType() != ColumnType.TEXT)
			throw new ISQLException("LIKE("+like+") needs a text expression, not "+valTest);
		Pattern pattern = Pattern.compile(like);
		if(valTest instanceof Array)
		{
			WhereView viewTest = WhereView.fromExpression(test);			
			WhereView viewRes = WhereView.none(viewTest.getTableNames());
			String[] arrTest = (String[]) ((Array)valTest).getArray();
			for(int i = 0; i < arrTest.length; i++)
				if(BooleanHelp.xor(matches(pattern, arrTest[i]), !negated))
					viewRes.addRowIndex(i);
			return viewRes;
		}
		else if(valTest instanceof IConstant)
		{
			String txtTest =  ((TextConstant)valTest).getValue();
			if(BooleanHelp.xor(matches(pattern, txtTest), !negated))
				return WhereView.fromColumns(values);
			else
			{
				String[] tables = WhereView.getTablesFromColumnKeys(test.getInvolvedColumns());							
				return WhereView.none(tables);
			}
		}
		else
			throw new AssertionError("The expression type for LIKE("+like+") is not recognized "+valTest.getClass());		
	}
	/**
	 * Returns true if the text matches the regular expression 
	 * @param pattern the regular expression pattern
	 * @param txtTest the text to test against the regular expression
	 * @return true if it matches, false otherwise
	 */
	private static boolean matches(Pattern pattern, String txtTest)
	{
		return pattern.matcher(txtTest).matches();
	}
	
	@Override
	public String toString()
	{
		String s = test.toString()+" LIKE "+like;
		if(negated)
			return "NOT "+s;
		else
			return s;
	}
}
