/*
 * Copyright (C) 2007 Andrea Ferrandi <aferrandi@users.sourceforge.net>. 
 * Released under the terms of the GNU LGPL v3.0.
 */
package isql.predicate;

import isql.ColumnKey;
import isql.ColumnValues;
import isql.IExpression;
import isql.ISQLException;
import isql.array.Array;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;


/**
 * Contains the tables involved in a predicate (a section of the where clause)
 * with the indices of their rows that are resulting from the predicate filter.
 * For example the predicate: 
 * 
 * 	user.name = 'brian'
 * 
 * will return a view containing, for the table user, the indices of all the rows that have the field
 * name content equal to the text "brian".
 * 
 * You can see the content of the object as a view in the database, where instead
 * to have all the values that have been filtered and joined for the view, you
 * have the indices of the tables rows
 * 
 * @author SHZ Oct 12, 2007
 */
public class WhereView
{
	/**
	 * since WhereView is equivalent to a VIEW in the database,
	 * the row of the view must have values for each involved table, 
	 * even if the table rows don't match the filter.
	 * In this case we say that the table does not have value for that row
	 * (equivalent to the NULL value in the database) 
	 */
	public static final Integer NO_ROW = Integer.valueOf(-1);
	
	/**
	 * map table -> indices that are matching the filter (or NO_ROW)
	 */
	private Map<String, List<Integer>> mapView = null;
	
	private WhereView(Map<String, List<Integer>> mapView)
	{
		this.mapView = mapView;
	}
	
	/**
	 * Builds the view from the column values.
	 * Remember that at the start of the evaluation the column values can have different
	 * size, since they can be seen as different tables (which have different length) 
	 * @param values the column values
	 * @return the view containing the tables indexes
	 */
	public static WhereView fromColumns(ColumnValues values)
	{
		Map<String, List<Integer>> mapAll = new TreeMap<String, List<Integer>>();
		for(Map.Entry<ColumnKey, Array> entry : values.entrySet())
		{
			String table = entry.getKey().getTableName();
			if(!mapAll.containsKey(table))
			{
				List<Integer> lstIndex = new ArrayList<Integer>();
				for(int i = 0; i < entry.getValue().getSize(); i++)
					lstIndex.add(i);
				mapAll.put(table, lstIndex);
			}
		}
		return new WhereView(mapAll);
	}

	/**
	 * Builds a view that, for all tables, does not contain any row index 
	 * @param arrTable the tables
	 * @return the view
	 */
	public static WhereView none(String[] arrTable)
	{
		TreeMap<String, List<Integer>> mapNone = new TreeMap<String, List<Integer>>();
		for(String table : arrTable)
			mapNone.put(table, new ArrayList<Integer>());
		return new WhereView(mapNone);
	}
	
	/**
	 * Builds a view from an expression (the expression contains the involved columns, and therefore
	 * the involved tables).
	 * The view is just a set of table without row indexes
	 * @param exp the expression
	 * @return the view
	 */
	public static WhereView fromExpression(IExpression exp)
	{
		Map<String, List<Integer>> mapRet = new TreeMap<String, List<Integer>>();
		for(ColumnKey key : exp.getInvolvedColumns())
		{
			String tableName = key.getTableName();
			if(!mapRet.containsKey(tableName))
				mapRet.put(tableName, new ArrayList<Integer>());
		}
		return new WhereView(mapRet);		
	}
	/**
	 * Adds the content of another view to this one
	 * @param otherView the other view 
	 */
	public void addAll(WhereView otherView)
	{
		mapView.putAll(otherView.mapView);
	}
	
	/**
	 * Adds a single row index for each table of the view
	 * @param index the row index
	 */
	public void addRowIndex(int index)
	{
		for(List<Integer> lstIndex : mapView.values())
			lstIndex.add(index);
	}
	
	/**
	 * Adds a table to the view with all the indices 
	 * @param table
	 * @param lstRowIndices
	 */
	public void addTableWithRowIndices(String table, List<Integer> lstRowIndices)
	{
		mapView.put(table, lstRowIndices);
	}
	/**
	 * Returns the indices for one specific table in this view
	 * @param table the table to get the indices for
	 * @return the indicess
	 */
	public List<Integer> getIndicesInTable(String table)
	{
		return mapView.get(table);
	}
	/**
	 * add tables without any index
	 * @param arrTable
	 */
	public void addEmptyTables(String[] arrTable)
	{
		for(String table : arrTable)
			addEmptyTable(table);
	}
	
	/**
	 * Retursn the names of the tables that are included in the view
	 * @return an array with the tables names
	 */
	public String[] getTableNames()
	{
		Set<String> tableNameSet = mapView.keySet();
		return tableNameSet.toArray(new String[tableNameSet.size()]);
	}
	
	/**
	 * Adds a table (if it does not already exist) in the view, without any index
	 * @param table the name of the table to add
	 */
	public void addEmptyTable(String table)
	{
		if(!mapView.containsKey(table))
			mapView.put(table, new ArrayList<Integer>());
	}
	
	/**
	 * Checks if a table already exists in the view
	 * @param table the name of the table to check
	 * @return true if the table exists, false otherwise
	 */
	public boolean hasTable(String table)
	{
		return mapView.containsKey(table);
	}
	
	/**
	 * column keys identifies columns and they are composed by the table and column name.
	 * This method extracts the table names for all columns 
	 * @param colKey a collection of column keys
	 * @return the table names of the columns, without repetition
	 */
	public static String[] getTablesFromColumnKeys(Collection<ColumnKey> colKey)
	{
		Set<String> setTableName = new HashSet<String>();
		for(ColumnKey key : colKey)
			setTableName.add(key.getTableName());
		return setTableName.toArray(new String[setTableName.size()]);
	}
	/**
	 * Since all the tables have by definition the same number of rows (possibly empty)
	 * the number of rows of the view is the same as the number of rows of the first
	 * table
	 * @return the number of rows in the view
	 * @throws ISQLException 
	 */
	public int getRowsNumber() throws ISQLException
	{
		if(mapView.size() > 0)
			return mapView.values().iterator().next().size();
		else
			throw new ISQLException("No tables in the view");
	}
	
	
}
 